/* 
   Things, in their Essence,
   are not of this world.
*/

#include <sys/time.h>

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include <math.h>
#include <string.h>

#include <pthread.h>
#include <sched.h>
#include <sys/mman.h>

#include "fweelin_block.h"
#include "fweelin_core.h"

VorbisDecoder::VorbisDecoder(Fweelin *app) : app(app), infd(0) {};
VorbisDecoder::~VorbisDecoder() {};

// Returns nonzero on error
int VorbisDecoder::ReadFromFile(FILE *in) {
  if (ov_open(in, &vf, NULL, 0) < 0) {
    printf("DISK: (VorbisFile) Input does not appear to be an Ogg "
	   "bitstream.\n");
    return 1;
  }

  vorbis_info *vi = ov_info(&vf,-1);

  if (vi->channels == 2) {
    printf("DISK: (VorbisFile) Stereo loop\n");
    stereo = 1;
  }
  else if (vi->channels == 1) {
    printf("DISK: (VorbisFile) Mono loop\n");
    stereo = 0;
  }
  else {
    printf("DISK: (VorbisFile) Unknown # of audio channels %d\n",vi->channels);
    return 1;
  }

  if (vi->rate != (signed int) app->getAUDIO()->get_srate()) {
    printf("DISK: (VorbisFile) Audio encoded at %dHz but we are running at "
	   "%dHz.\n"
	   "Samplerate conversion not yet supported.\n",
	   (int) vi->rate,app->getAUDIO()->get_srate());
    return 1;
  }
    
  infd = in;
  return 0;
};

int VorbisDecoder::Decode(float ***pcm_channels, nframes_t max_len) {
  return ov_read_float(&vf,pcm_channels,max_len,&current_section);
};

VorbisEncoder::VorbisEncoder(Fweelin *app, char stereo) : app(app), 
							  stereo(stereo),
							  outfd(0) {
  const static float VORBIS_ENCODE_QUALITY = 0.5;

  // Setup vorbis
  vorbis_info_init(&vi);
  if (vorbis_encode_init_vbr(&vi,(stereo ? 2 : 1),app->getAUDIO()->get_srate(),
			     VORBIS_ENCODE_QUALITY))
    return;
  
  // Comment
  vorbis_comment_init(&vc);
  vorbis_comment_add_tag(&vc,"ENCODER","FreeWheeling");
  
  // Analysis state/Aux encoding storage
  vorbis_analysis_init(&vd,&vi);
  vorbis_block_init(&vd,&vb);
  
  /* set up our packet->stream encoder */
  /* pick a random serial number; that way we can more likely build
     chained streams just by concatenation */
  srand(time(NULL));
  ogg_stream_init(&os,rand());
};

VorbisEncoder::~VorbisEncoder() {
  /* Vorbis clean up and exit.  vorbis_info_clear() must be called last */  
  ogg_stream_clear(&os);
  vorbis_block_clear(&vb);
  vorbis_dsp_clear(&vd);
  vorbis_comment_clear(&vc);
  vorbis_info_clear(&vi);
};

void VorbisEncoder::DumpToFile(FILE *out) {
  outfd = out;

  // Write vorbis header
  ogg_packet header;
  ogg_packet header_comm;
  ogg_packet header_code;
  
  vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
  ogg_stream_packetin(&os,&header); /* automatically placed in its own page */
  ogg_stream_packetin(&os,&header_comm);
  ogg_stream_packetin(&os,&header_code);
  
  while(1) {
    int result = ogg_stream_flush(&os,&og);
    if (result == 0)
      break;
    fwrite(og.header,1,og.header_len,outfd);
    fwrite(og.body,1,og.body_len,outfd);
  }
}

// This analyzes and dumps any remaining frames out to file
long int VorbisEncoder::Encode() {
  if (outfd != 0) {
    long int outputsize = 0;
    
    // Now do the real analysis!! Meat n potatoes!
    while(vorbis_analysis_blockout(&vd,&vb) == 1) {
      /* analysis, assume we want to use bitrate management */
      vorbis_analysis(&vb,NULL);
      vorbis_bitrate_addblock(&vb);
      
      while(vorbis_bitrate_flushpacket(&vd,&op)) {
	/* weld the packet into the bitstream */
	ogg_stream_packetin(&os,&op);
	
	/* write out pages (if any) */
	char eos = 0;
	while (!eos) {
	  int result = ogg_stream_pageout(&os,&og);
	  if (result == 0)
	    break;
	  fwrite(og.header,1,og.header_len,outfd);
	  fwrite(og.body,1,og.body_len,outfd);
	  
	  // Add to the output size
	  outputsize += og.header_len;
	  outputsize += og.body_len;
	  
	  if (ogg_page_eos(&og))
	    eos = 1;
	}
      }
    }

    return outputsize;
  } else {
    printf("BLOCK: ERROR: OGG encoder finished but Encode() called!\n");
    return 0;
  }
}

BlockReadManager::BlockReadManager(FILE *in, AutoReadControl *arc, 
				   BlockManager *bmg, 
				   nframes_t peaksavgs_chunksize) :
  ManagedChain(0,0), in(in), arc(arc), dec(0), bmg(bmg), pa_mgr(0),
  peaksavgs_chunksize(peaksavgs_chunksize)
{
  pthread_mutex_init(&decode_lock,0);
};

BlockReadManager::~BlockReadManager() {
  if (in != 0) {
    End(0);
  }

  pthread_mutex_destroy (&decode_lock);
};

// Start decoding from the given file
void BlockReadManager::Start(FILE *new_in) {
  if (dec != 0)
    return; // Already running

  if (in == 0 && new_in != 0) {
    // New input specified
    in = new_in;
  }

  if (in != 0) {
    // Start vorbis decoder
    dec = new VorbisDecoder(bmg->GetApp());
    b = (AudioBlock *) bmg->GetApp()->getPRE_AUDIOBLOCK()->RTNew();
    i = new AudioBlockIterator(b,DECODE_CHUNKSIZE,
			       bmg->GetApp()->getPRE_EXTRACHANNEL());
    
    // Compute audio peaks & averages for display
    if (peaksavgs_chunksize != 0) {
      AudioBlock *peaks = (AudioBlock *) b->RTNew(),
	*avgs = (AudioBlock *) b->RTNew();
      if (peaks == 0 || avgs == 0) {
	printf("BlockReadManager: ERROR: No free blocks for peaks/avgs\n");
	if (peaks != 0)
	  peaks->RTDelete();
	if (avgs != 0)
	  avgs->RTDelete();
	pa_mgr = 0;
      } else {
	b->AddExtendedData(new BED_PeaksAvgs(peaks,avgs,peaksavgs_chunksize));
	pa_mgr = bmg->PeakAvgOn(b,i,1);
      }
    }
    
    // Tell decoder to read from file
    if (dec->ReadFromFile(in))
      // End in error, freeing vorbis 
      End(1);
  }
};

void BlockReadManager::End(char error) {
  // Can't end while decoding- lock mutex
  pthread_mutex_lock (&decode_lock);

  // Stop decoder
  if (dec != 0) {
    dec->Stop();
    delete dec;

    // Chop block to current position
    i->EndChain();

    // End peaks avgs compute now
    if (pa_mgr != 0) {
      // Catchup then end
      pa_mgr->Manage();
      pa_mgr->End();
      bmg->PeakAvgOff(b);
      pa_mgr = 0;
    }

    delete i;
    i = 0;
    dec = 0;
  }

  if (in != 0) {
    printf("DISK: Close input.\n");
    // Vorbis decoder closes file through ov_clear!
    in = 0;
  }

  if (b != 0) {
    if (error) {
      // If error, delete chain and send zero to ReadComplete
      b->DeleteChain();
      b = 0;
    }
    
    // Callback
    if (arc != 0)
      arc->ReadComplete(b);

    b = 0;
  }

  pthread_mutex_unlock(&decode_lock);
};

int BlockReadManager::Manage() {
  if (in == 0) {
    // Not currently decoding
    if (arc != 0) {
      // We have a callback, get a chain to load
      arc->GetReadBlock(&in);
      
      if (in != 0)
	// We got a chain to load, so begin
	Start();
    } else {
      // No callback, so we are done!
      return 1;
    }
  }

  if (in != 0) {
    // Make sure we have started on this blockchain
    if (dec == 0)
      Start();

    // Continue decoding
    pthread_mutex_lock (&decode_lock);

    // Now that we have the lock, make sure we are still active
    if (in != 0) {
      float **inbuf;
      char stereo = dec->IsStereo();

      for (int pass = 0; in != 0 && pass < NUM_DECODE_PASSES; pass++) {
	// Make sure we have an extra block in our chain
	if (i->GetCurBlock()->next == 0) {
	  AudioBlock *nw = (AudioBlock *) i->GetCurBlock()->RTNew();
	  if (nw == 0) {
	    // Pause decode until block is here
	    printf("BlockReadManager: Waiting for block.\n");
	    pthread_mutex_unlock (&decode_lock);
	    return 0;
	  }
	  
	  i->GetCurBlock()->Link(nw);
	}
	
	// Decode samples
	int len = dec->Decode(&inbuf,DECODE_CHUNKSIZE);
	if (len > 0) {
	  // Store in block actual samples read, waiting for alloc if necessary
	  if (stereo) 
	    i->PutFragment(inbuf[0],inbuf[1],len,1);
	  else
	    i->PutFragment(inbuf[0],0,len,1);
	  
	  i->NextFragment();
	} else if (len == 0) {
	  // EOF
	  pthread_mutex_unlock (&decode_lock);
	  End(0);
	} else {
	  // Stream error- just continue
	  printf("BlockReadManager (VorbisFile): OGG stream error!\n");
	}
      }
    }

    pthread_mutex_unlock (&decode_lock);
  }

  return 0;
};

BlockWriteManager::BlockWriteManager(FILE *out, AutoWriteControl *awc, 
				     BlockManager *bmg, AudioBlock *b, 
				     AudioBlockIterator *i) : 
  ManagedChain(b,i), out(out), len(0), awc(awc), enc(0), bmg(bmg)
{
  pthread_mutex_init(&encode_lock,0);
};

BlockWriteManager::~BlockWriteManager() {
  if (b != 0) {
    End();
  }

  pthread_mutex_destroy (&encode_lock);
};

// Start encoding the given block chain to the given file
void BlockWriteManager::Start(FILE *new_out, AudioBlock *new_b, 
			      AudioBlockIterator *new_i,
			      nframes_t new_len) {
  if (enc != 0)
    return; // Already running

  if (b == 0 && new_b != 0) {
    // New chain specified
    // printf("start: new chain b: %p new_b: %p\n",b,new_b);
    b = new_b;
    i = new_i;
    out = new_out;
    len = new_len;
  }

  if (b != 0) {
    //printf("start: out: %p new_out: %p, b: %p new_b: %p\n",out,new_out,
    //   b,new_b);

    // Start vorbis encoder
    enc = new VorbisEncoder(bmg->GetApp(),b->IsStereo());
    ei = new AudioBlockIterator(b,ENCODE_CHUNKSIZE);
    
    // Set encoder to dump to file
    enc->DumpToFile(out);
  }
};

void BlockWriteManager::End() {
  // Can't end while encoding- lock mutex
  pthread_mutex_lock (&encode_lock);

  // Stop encoder
  if (enc != 0) {
    enc->Stop();
    delete enc;
    delete ei;
    enc = 0;
  }

  if (out != 0) {
    printf("DISK: Close output.\n");
    fclose(out);
    out = 0;
  }

  b = 0;
  i = 0;
  len = 0;
  
  pthread_mutex_unlock(&encode_lock);
};

int BlockWriteManager::Manage() {
  if (b == 0) {
    // Not currently encoding
    if (awc != 0) {
      // We have a callback, get a chain to save
      awc->GetWriteBlock(&out,&b,&i,&len);
      
      if (b != 0)
	// We got a chain to save, so begin
	Start();
    } else {
      // No callback, so we are done!
      return 1;
    }
  }

  if (b != 0) {
    // Make sure we have started on this blockchain
    if (enc == 0)
      Start();

    // Continue encoding
    pthread_mutex_lock (&encode_lock);

    // Now that we have the lock, make sure we are still active
    if (b != 0) {
      nframes_t pos = ei->GetTotalLength2Cur(),
	remaining = len-pos;
      nframes_t num = MIN(ENCODE_CHUNKSIZE,remaining);
      
      float **obuf = enc->GetAnalysisBuffer(num);    
      sample_t *ibuf[2];
      if (enc->IsStereo()) {
	// Stereo
	ei->GetFragment(&ibuf[0],&ibuf[1]);
	memcpy(obuf[0],ibuf[0],sizeof(float) * num);
	memcpy(obuf[1],ibuf[1],sizeof(float) * num);
      } else {
	// Mono
	ei->GetFragment(&ibuf[0],0);
	memcpy(obuf[0],ibuf[0],sizeof(float) * num);
      }

      enc->WroteToBuffer(num);
      enc->Encode();
      
      if (remaining <= ENCODE_CHUNKSIZE) {
	// Finished encoding
	pthread_mutex_unlock (&encode_lock);
	End();
      }
      else
	ei->NextFragment();
    }

    pthread_mutex_unlock (&encode_lock);
  }

  return 0;
};

BED_PeaksAvgs::~BED_PeaksAvgs() {
  peaks->DeleteChain();
  avgs->DeleteChain();
};

int BED_MarkerPoints::CountMarkers() {
  TimeMarker *cur = markers;
  int markcnt = 0;
  while (cur != 0) {
    markcnt++;
    cur = cur->next;
  };
  
  return markcnt;
};

BED_MarkerPoints::~BED_MarkerPoints() {
  TimeMarker *cur = markers;
  while (cur != 0) {
    TimeMarker *tmp = cur->next;
    cur->RTDelete();
    cur = tmp;
  } 
};

// Returns the nth marker before (in time) the current offset passed
// This method correctly handles a reverse wrap case when the
// marker NBeforeCur is farther ahead in the attached block than the
// current offset
TimeMarker *BED_MarkerPoints::GetMarkerNBeforeCur(int n, nframes_t curofs) {
  TimeMarker *cur = markers;
  signed int markcnt = 0;
  int totalmarks = CountMarkers();
  
  if (totalmarks == 0) 
    return 0; // No solution because there are no markers!
  else {
    while (cur != 0 && cur->markofs < curofs) {
      markcnt++;
      cur = cur->next;
    }
    
    // Markcnt now indexes the next marker after curofs
    // Go N before!
    markcnt -= n;
    while (markcnt < 0) {
      // Reverse wrap case
      markcnt += totalmarks;
    }
    
    // Now get marker indexed by markcnt
    cur = markers;
    int markcnt2 = 0;
    while (cur != 0 && markcnt2 != markcnt) {
      markcnt2++;
      cur = cur->next;
    }
    
    // & return it
    return cur;
  }
};

// Create a new extra channel
BED_ExtraChannel::BED_ExtraChannel(nframes_t len) {
  if (len > 0) {
    //printf("allocating origbuf..\n");
    origbuf = buf = new sample_t[len];
    //printf("done: origbuf %p: buf %p!..\n",origbuf,buf);
  }
  else
    origbuf = buf = 0;
};

BED_ExtraChannel::~BED_ExtraChannel() {
  if (origbuf != 0) {
    //printf("deleting origbuf %p: buf %p!..\n",origbuf,buf);
    delete[] origbuf;
    //printf("done!\n");
  }
};

// Create a new audioblock as the beginning of a block list
AudioBlock::AudioBlock(nframes_t len) : len(len), next(0),
					first(this), xt(0)
{
  if (len > 0)
    origbuf = buf = new sample_t[len];
}

// Create a new audioblock and link up the specified block to it
AudioBlock::AudioBlock(AudioBlock *prev, nframes_t len) : 
  len(len), next(0), first(prev->first), xt(0) {
  prev->next = this;
  if (len > 0)
    origbuf = buf = new sample_t[len];
}

AudioBlock::~AudioBlock() {
  //printf("~AudioBlock len: %d dsz: %d\n",len,sizeof(*buf));
  if (origbuf != 0)
    delete[] origbuf;
}

// Clears this block chain- not changing length
void AudioBlock::Zero() {
  AudioBlock *cur = first;
  while (cur != 0) {
    memset(cur->buf,0,sizeof(sample_t)*cur->len);
    cur = cur->next;
  }
};

// Link us up to the specified block
void AudioBlock::Link(AudioBlock *to) {
  to->first = first;
  to->next = 0;
  next = to;
}

// RT safe
void AudioBlock::ChopChain() {
  // Chop off the chain at this block
  // Freeing unused blocks
  AudioBlock *cur = next;
  next = 0;
  while (cur != 0) {
    AudioBlock *tmp = cur->next;
    if (cur->xt != 0)
      printf("BLOCK: WARNING: XT data not freed in ChopChain! "
	     "Possible leak.\n");

    cur->RTDelete();
    cur = tmp;
  }
}

void AudioBlock::SmoothEnd() {
  // Smooth the end of this chain into the beginning for looping
  AudioBlock *smoothblk;
  nframes_t smoothofs,
    smoothcnt,
    totallen = GetTotalLen();
  if (totallen >= AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN)
    smoothcnt = totallen - AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN;
  else
    return; // Very short chain, can't smooth it

  // Get block & offset for start of smooth
  SetPtrsFromAbsOffset(&smoothblk, &smoothofs, smoothcnt);
  if (smoothblk == 0) {
    // Shouldn't happen
    printf("AudioBlock: ERROR: Block position mismatch in SmoothEnd\n");
    exit(1);
  }

  AudioBlock *startblk = first;
  nframes_t startofs = 0;

  // Second channel?
  BED_ExtraChannel *rightstartblk = 
    (BED_ExtraChannel *) startblk->GetExtendedData(T_BED_ExtraChannel),
    *rightsmoothblk = (BED_ExtraChannel *)
    smoothblk->GetExtendedData(T_BED_ExtraChannel);
  char stereo = 0;
  if (rightstartblk != 0 && rightsmoothblk != 0)
    stereo = 1; 
    
  float mix = 0.0,
    dmix = 1./(float)AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN;
  for (nframes_t i = 0; i < AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN; i++, 
	 mix += dmix) {
    smoothblk->buf[smoothofs] = mix*startblk->buf[startofs] + 
      (1.0-mix)*smoothblk->buf[smoothofs];
    if (stereo)
      rightsmoothblk->buf[smoothofs] = mix*rightstartblk->buf[startofs] + 
	(1.0-mix)*rightsmoothblk->buf[smoothofs];
      
    smoothofs++;
    startofs++;
    if (smoothofs >= smoothblk->len) {
      smoothblk = smoothblk->next;
      smoothofs = 0;
      if (smoothblk == 0) {
	if (i+1 < AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN) {
	  // Shouldn't happen
	  printf("AudioBlock: ERROR: (smoothblk) Block size mismatch in "
		 "SmoothEnd: i: %d\n", i);
	  exit(1);
	}
	
	rightsmoothblk = 0;
      } else {
	rightsmoothblk = (BED_ExtraChannel *)
	  smoothblk->GetExtendedData(T_BED_ExtraChannel);
	if (stereo && rightsmoothblk == 0) {
	  // Shouldn't happen
	  printf("AudioBlock: ERROR: (smoothblk) Right channel "
		 "disappeared!\n");
	  exit(1);
	}
      }
    }
    if (startofs >= startblk->len) {
      startblk = startblk->next;
      startofs = 0;
      if (startblk == 0) {
	if (i+1 < AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN) {
	  // Shouldn't happen
	  printf("AudioBlock: ERROR: (startblk) Block size mismatch in "
		 "SmoothEnd\n");
	  exit(1);
	}
	
	rightstartblk = 0;
      } else {
	rightstartblk = (BED_ExtraChannel *)
	  startblk->GetExtendedData(T_BED_ExtraChannel);
	if (stereo && rightstartblk == 0) {
	  // Shouldn't happen
	  printf("AudioBlock: ERROR: (startblk) Right channel disappeared!\n");
	  exit(1);
	}
      }
    }
  }

  // Adjust first buf to skip samples that were smoothed into end
  if (first->len < AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN) {
    printf("AudioBlock: WARNING: First block is very short, "
	   "no adjust in SmoothEnd.\n");
  } else {
    first->buf += AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN;
    first->len -= AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN;

    BED_ExtraChannel *rightfirst = 
      (BED_ExtraChannel *) first->GetExtendedData(T_BED_ExtraChannel);
    if (rightfirst != 0) 
      rightfirst->buf += AUDIOBLOCK_SMOOTH_ENDPOINTS_LEN;
  }

  //printf("endsmooth: startofs: %d smoothofs: %d len: %d\n",
  //	 startofs, smoothofs, totallen);
}

// not RT safe
void AudioBlock::DeleteChain() {
  AudioBlock *cur = first;
  while (cur != 0) {
    // First erase any extended data
    BlockExtendedData *curxt = cur->xt;
    while (curxt != 0) {
      BlockExtendedData *tmpxt = curxt->next;
      delete curxt;
      curxt = tmpxt;
    }

    // Then the block itself!
    AudioBlock *tmp = cur->next;
    cur->RTDelete();
    cur = tmp;
  }
}

// Returns the total length of this audio chain
nframes_t AudioBlock::GetTotalLen() {
  nframes_t tally = 0;
  AudioBlock *cur = first;
  while (cur != 0) {
    tally += cur->len;
    cur = cur->next;
  }
  
  return tally;
};

// Gets extended data for this block
BlockExtendedData *AudioBlock::GetExtendedData(BlockExtendedDataType x) { 
  BlockExtendedData *cur = xt;
  while (cur!=0 && cur->GetType()!=x)
    cur = cur->next;
  return cur;
};

// Add this extended data to the list of extended data for this block
void AudioBlock::AddExtendedData(BlockExtendedData *nw) {
  nw->next = xt;
  xt = nw;
};

// Finds the audioblock and offset into that block that correspond
// to the provided absolute offset into this chain 
void AudioBlock::SetPtrsFromAbsOffset(AudioBlock **ptr, nframes_t *blkofs, 
				      nframes_t absofs) {
  AudioBlock *cur = first,
    *prev = 0;
  nframes_t curofs = 0,
    prevofs = 0;
  while (cur != 0 && curofs <= absofs) {
    prevofs = curofs;
    curofs += cur->len;
    prev = cur;
    cur = cur->next;
  }  
  
  if (curofs < absofs || prev == 0) {
    // We're at the end and we still haven't traversed far enough!
    // Return err
    *ptr = 0;
    return;
  }
  
  *ptr = prev;
  *blkofs = absofs-prevofs;
};

// Generates a subchain of AudioBlocks by copying the samples between offsets
// from & to.. offsets expressed absolutely with reference to
// beginning of this chain!
// Samples are copied up to but not including toofs
// If toofs < fromofs, generates a subblock that includes the end of
// the block and the beginning up to toofs (wrap case)
// Returns the last block in the new subchain
// Flag sets whether to copy stereo channel if there is one
// Realtime safe?
AudioBlock *AudioBlock::GenerateSubChain(nframes_t fromofs, nframes_t toofs,
					 char copystereo) {
  if (toofs == fromofs)
    return 0;
  // Handle wrap case
  if (toofs < fromofs)
    toofs += GetTotalLen();

  // Stereo?
  BED_ExtraChannel *rightfirst = (BED_ExtraChannel *)
    first->GetExtendedData(T_BED_ExtraChannel);
  char stereo = (copystereo && rightfirst != 0);
    
  // Get enough new blocks to fit the subblock size
  nframes_t sublen = toofs - fromofs,
    suballoc = 0;
  AudioBlock *subfirst = 0,
    *subcur = 0;
  while (suballoc < sublen) {
    if (subfirst == 0) {
      subcur = subfirst = (AudioBlock *)RTNew();
      if (subfirst == 0) {
	printf("Err: GenerateSubChain- No new blocks available\n");
	return 0;
      }
    }
    else {
      AudioBlock *tmp = (AudioBlock *)RTNew();
      if (tmp == 0) {
	printf("Err: GenerateSubChain- No new blocks available\n");
	return 0;
      }
      subcur->Link(tmp);
      subcur = subcur->next;
    }

    if (stereo) {
      BED_ExtraChannel *rightsub = (BED_ExtraChannel*)rightfirst->RTNew(); 
      if (rightsub == 0) {
	printf("Err: GenerateSubChain- No new right blocks available\n");
	return 0;
      }
      subcur->AddExtendedData(rightsub);
    }

    suballoc += subcur->len;
  }
    
  // Find starting pos in chain
  AudioBlock *cur;
  nframes_t curofs;
  SetPtrsFromAbsOffset(&cur,&curofs,fromofs);

  subcur = subfirst;
  nframes_t subofs = 0,
    remaining = sublen; 

  // Extra channel
  BED_ExtraChannel *subright = 
    (stereo ? (BED_ExtraChannel *) subcur->GetExtendedData(T_BED_ExtraChannel)
     : 0),
    *right = 
    (stereo ? (BED_ExtraChannel *) cur->GetExtendedData(T_BED_ExtraChannel)
     : 0);

  do {
    nframes_t n = MIN(cur->len-curofs,remaining);
    n = MIN(n,subcur->len-subofs);

    memcpy(&subcur->buf[subofs],&cur->buf[curofs],n*sizeof(sample_t));
    if (stereo)
      memcpy(&subright->buf[subofs],&right->buf[curofs],n*sizeof(sample_t));
      
    subofs += n;
    curofs += n;
    remaining -= n;
    if (curofs >= cur->len) {
      cur = cur->next;
      curofs = 0;
      if (cur == 0)
	// Past the end of the block chain-- wrap around to first
	cur = first;

      if (stereo)
	right = (BED_ExtraChannel *) cur->GetExtendedData(T_BED_ExtraChannel);
    }

    if (subofs >= subcur->len) {
      subcur = subcur->next;
      subofs = 0;
      if (subcur == 0 && remaining) {
	// This should never happen, because we make the destination
	// chain long enough to hold the subblock
	printf("Err: GenerateSubChain destination block size mismatch\n");
	exit(1);
      }

      if (stereo)
	subright = (BED_ExtraChannel *) 
	  subcur->GetExtendedData(T_BED_ExtraChannel);
    }
  } while (remaining);

  // Truncate the last presized subblock to the right length
  subcur->len = subofs;

  // And return the last block in the chain
  return subcur;
};

// *** To be tested! ***
//
// Removes the last 'hacklen' samples from this block chain
// Returns nonzero on error!
// Realtime safe?
int AudioBlock::HackTotalLengthBy(nframes_t hacklen) {
  // Compute the position of hack
  nframes_t chainlen = GetTotalLen();
  //printf("Hack (b): %ld\n",chainlen);
  AudioBlock *hackblk;
  nframes_t hackofs;
  if (chainlen <= hacklen)
    return 1;
  SetPtrsFromAbsOffset(&hackblk,&hackofs,chainlen-hacklen);
  
  // Now hackblk[hackofs] should be the first sample to erase at the
  // end of the chain
  hackblk->len = hackofs; // Truncate length of block
  // Note: Block memory isnt resized, so extra memory lingers until
  // the block is erased
  
  AudioBlock *tmp = hackblk->next;
  hackblk->next = 0; // End of chain at hack position
  
  // Erase remaining blocks in chain
  hackblk = tmp; 
  while (hackblk != 0) {
    tmp = hackblk->next;
    hackblk->RTDelete();
    hackblk = tmp;
  }

  // **** To be implemented
  // **** Hack stereo samples

  //printf("Hack (e): %ld\n",GetTotalLen());
  return 0;
};

// Inserts the new blockchain at the beginning of this block chain
// Returns a pointer to the new first block
AudioBlock *AudioBlock::InsertFirst(AudioBlock *nw) {
  // Move to the end of the passed chain
  while (nw->next != 0)
    nw = nw->next;

  // Link new block to the beginning of our chain
  nw->next = first;
  
  // Link all first pointers to first block in new chain
  AudioBlock *cur = first;
  while (cur != 0) {
    cur->first = nw->first;
    cur = cur->next;
  }
  
  return nw->first;
};

AudioBlockIterator::AudioBlockIterator(AudioBlock *firstblock,  
				       nframes_t fragmentsize,
				       PreallocatedType *pre_extrachannel) :
  pre_extrachannel(pre_extrachannel), currightblock(0), nextrightblock(0),
  curblock(firstblock), nextblock(0), curblkofs(0), nextblkofs(0), curcnt(0),
  nextcnt(0), fragmentsize(fragmentsize), stopped(0) {
  fragment[0] = new sample_t[fragmentsize];
  fragment[1] = new sample_t[fragmentsize];
}

AudioBlockIterator::~AudioBlockIterator() { 
  delete[] fragment[0];
  delete[] fragment[1];
};

// Stores in cnt the absolute count corresponding to the given
// block and offset
void AudioBlockIterator::GenCnt(AudioBlock *blk, nframes_t blkofs, 
				nframes_t *cnt) {
  AudioBlock *cur = curblock->first;
  nframes_t curofs = 0;
  while (cur != 0 && cur != blk) {
    // Add the length of this whole block
    curofs += cur->len;
    cur = cur->next;
  }
  
  if (cur == 0) {
    // Given block not found!
    *cnt = 0;
  }
  else {
    // We are now on the given block- add only current block offset
    curofs += blkofs;
    *cnt = curofs;
  }
}

void AudioBlockIterator::Jump(nframes_t ofs) {
  // Quantize the specified offset to within the limits of the blockchain
  ofs = ofs % curblock->GetTotalLen();

  curblock->SetPtrsFromAbsOffset(&curblock,&curblkofs,ofs);
  if (curblock == 0) {
    printf("Err: AudioBlockIterator::Jump- Pointer/size mismatch\n");
    exit(1);
  }
  currightblock = 0;
  nextrightblock = 0;
  nextblock = 0;
  nextblkofs = 0;
  curcnt = ofs;
  nextcnt = 0;
}

void AudioBlockIterator::GenConstants() {
  // Need to recompute curcnt to account for extra blocks?
  GenCnt(curblock,curblkofs,&curcnt);
  GenCnt(nextblock,nextblkofs,&nextcnt);
}

// Moves iterator to start position
void AudioBlockIterator::Zero() {
  currightblock = 0;
  nextrightblock = 0;
  curblock = curblock->first;	  
  curblkofs = 0; 
  curcnt = 0;
  nextblock = 0;
}

// Advances to the next fragment
void AudioBlockIterator::NextFragment() {
  // Only advance if not stopped
  if (!stopped) {
    // We must first get a fragment before advancing
    if (nextblock == 0)
      GetFragment(0,0);
    
    currightblock = nextrightblock;
    nextrightblock = 0;
    curblock = nextblock;
    curblkofs = nextblkofs;
    curcnt = nextcnt;
    
    nextblock = 0;
  }
}

// PutFragment puts the specified fragment back into this AudioBlock
// returns nonzero if the end of the block is reached, and we wrap to next
int AudioBlockIterator::PutFragment (sample_t *frag_l, sample_t *frag_r,
				     nframes_t size_override, 
				     char wait_alloc) {
  nframes_t fragofs = 0;
  nframes_t n = (size_override == 0 ? fragmentsize : size_override);
  int wrap = 0;
  
  // Keep local track of pointers
  // Since we have to be threadsafe
  BED_ExtraChannel *lclrightblock = currightblock;
  AudioBlock *lclblock = curblock;
  nframes_t lclblkofs = curblkofs,
    lclcnt = curcnt;

  nframes_t nextbit;
  do {
    nextbit = MIN((nframes_t) n, lclblock->len - lclblkofs);

    // Copy back into our block

    // Left
    memcpy(&lclblock->buf[lclblkofs],
	   &frag_l[fragofs],
	   sizeof(sample_t)*nextbit);

    // If right channel is given and we don't know the right block, find it
    if (frag_r != 0) {
      if (lclrightblock == 0) {
	lclrightblock = (BED_ExtraChannel *) 
	  lclblock->GetExtendedData(T_BED_ExtraChannel);
	if (lclrightblock == 0) {
	  // No right channel exists but we are being told to put data there--
	  // so create a right channel!
	  if (pre_extrachannel == 0) {
	    printf("BLOCK: ERROR: Need to make right channel buffer but no "
		   "preallocator was passed!\n");
	    return 0;
	  } else {
	    do {
	      lclrightblock = (BED_ExtraChannel *) pre_extrachannel->RTNew();
	      if (lclrightblock == 0 && wait_alloc) {
		printf("BLOCK: Waiting for BED_ExtraChannel.\n");
		usleep(10000); // Wait then try again
	      }
	    } while (lclrightblock == 0 && wait_alloc);
	    if (lclrightblock != 0)
	      lclblock->AddExtendedData(lclrightblock);
	    else {
	      printf("BLOCK: ERROR: RTNew() failed for BED_ExtraChannel.\n");
	      return 0;
	    }
	  }
	}
      }
      
      // Right
      memcpy(&lclrightblock->buf[lclblkofs],
	     &frag_r[fragofs],
	     sizeof(sample_t)*nextbit);
    }
    
    fragofs += nextbit;
    lclblkofs += nextbit;
    lclcnt += nextbit;
    n -= nextbit;
    
    if (lclblkofs >= lclblock->len) { // if (n) is wrong
      // If we get here, it means this block has been fully dumped
      // so we need the next block
      
      wrap = 1;
      if (lclblock->next == 0) {
	// END OF AUDIOBLOCK LIST, LOOP TO BEGINNING
	lclblock = lclblock->first;
	lclblkofs = 0;
	lclcnt = lclblkofs;
        lclrightblock = 0;
      }
      else {
	lclblock = lclblock->next;
	lclblkofs = 0; 
	lclrightblock = 0;
      }
    }
  } while (n);
  
  nextblock = lclblock,
    nextrightblock = lclrightblock,
    nextblkofs = lclblkofs,
    nextcnt = lclcnt;
  
  return wrap;
}

// Returns the current fragment of audio
// nextblock and nextblkofs become the new block pointer and offset
// for the next fragment
void AudioBlockIterator::GetFragment(sample_t **frag_l, sample_t **frag_r) {
  nframes_t fragofs = 0;
  nframes_t n = fragmentsize;
  
  // Keep local track of pointers
  // Since we have to be threadsafe
  BED_ExtraChannel *lclrightblock = currightblock;
  AudioBlock *lclblock = curblock;
  nframes_t lclblkofs = curblkofs,
    lclcnt = curcnt;
  
  nframes_t nextbit;
  do {
    nextbit = MIN((nframes_t) n, lclblock->len - lclblkofs);
      
    if (nextbit) {
      // Left
      memcpy(&fragment[0][fragofs],
	     &lclblock->buf[lclblkofs],
	     sizeof(sample_t)*nextbit);

      // If right channel is given and we don't know the right block, find it
      if (frag_r != 0) {
	if (lclrightblock == 0) {
	  lclrightblock = (BED_ExtraChannel *) 
	    lclblock->GetExtendedData(T_BED_ExtraChannel);
	  if (lclrightblock == 0) {
	    // No right channel exists but we are being told to get data--
	    printf("BLOCK: ERROR: Iterator asked for right channel but none "
		   "exists!\n");
	    return;
	  }
	}
	
	// Right
	memcpy(&fragment[1][fragofs],
	       &lclrightblock->buf[lclblkofs],
	       sizeof(sample_t)*nextbit);
      } else
	// Make -sure- we don't jump blocks and keep old rightblock
	lclrightblock = 0;

      fragofs += nextbit;
      lclblkofs += nextbit;
      lclcnt += nextbit;
      n -= nextbit;
    }
      
    if (lclblkofs >= lclblock->len) {
      // If we get here, it means this block is at an end
      // so we need the next block  
      if (lclblock->next == 0) {
	// END OF AUDIOBLOCK LIST, LOOP TO BEGINNING
	lclblock = lclblock->first;
	lclblkofs = 0;
	lclcnt = lclblkofs;
        lclrightblock = 0;
      } else {
	lclblock = lclblock->next;
	lclblkofs = 0; // Beginning of next block
        lclrightblock = 0;
      }
    }
  } while (n);
  
  nextblock = lclblock,
    nextrightblock = lclrightblock,
    nextblkofs = lclblkofs,
    nextcnt = lclcnt;

  // Return fragment buffers
  if (frag_l != 0)
    *frag_l = fragment[0];
  if (frag_r != 0) 
    *frag_r = fragment[1];
};

// RT safe
void AudioBlockIterator::EndChain() {
  // Mark iterator stopped
  stopped = 1;
  
  //printf("Crop: %ld to %ld\n", curblock->len, curblkofs);
  if (curblkofs < (nframes_t) fragmentsize) {
    //printf("WARNING: REALLY SHORT BLOCK: %d\n",curblkofs);
    curblock->len = fragmentsize;
  }
  else
    curblock->len = curblkofs;    
  
  // Stop the chain at this place
  curblock->ChopChain();
}

int GrowChainManager::Manage() {
  // Manage chain growth
  if (!i->IsStopped()) 
    if (i->GetCurBlock()->next == 0) {
      // Iterator is running and no more blocks at the end
      // of this chain! Get another and link it up.
      AudioBlock *nw = (AudioBlock *) i->GetCurBlock()->RTNew();
      if (nw == 0) {
	printf("GrowChainManager: ERROR: No free blocks to grow chain\n");
	return 0;
      }
      i->GetCurBlock()->Link(nw);

      // Second channel allocating is done as needed by PutFragment
      // in the iterator
    }

  return 0;
};

void PeaksAvgsManager::Setup() {
  // Setup iterators
  pa = (BED_PeaksAvgs *)(b->GetExtendedData(T_BED_PeaksAvgs));
  // To do: Optimize this to bypass iterator- 
  // it would be faster since we are moving one sample at a time!
  peaksi = new AudioBlockIterator(pa->peaks,1);
  avgsi = new AudioBlockIterator(pa->avgs,1);
  mi = new AudioBlockIterator(b,1);
  stereo = b->IsStereo();

  if (grow) {
    // Grow peaks & avgs blocks
    bmg->GrowChainOn(pa->peaks,peaksi);
    bmg->GrowChainOn(pa->avgs,avgsi);
  }
};

PeaksAvgsManager::~PeaksAvgsManager() {
  if (bmg != 0) {
    End();

    delete peaksi;
    delete avgsi;
    delete mi;
  }
};

void PeaksAvgsManager::End() {
  if (!ended) {
    ended = 1;

    if (grow) {
      // End chain at current pos!
      peaksi->EndChain();
      avgsi->EndChain();

      bmg->GrowChainOff(pa->peaks);
      bmg->GrowChainOff(pa->avgs);
    }
  }
}

int PeaksAvgsManager::Manage() {
  // Stop if ended..
  if (ended)
    return 1;

  // Compute running peaks and averages
  
  char wrap;
  do {
    wrap = 0;
    
    // Get current position in iterator 
    nframes_t curcnt = i->GetTotalLength2Cur();
    
    if (curcnt < lastcnt) {
      // We have a wrap condition!- loop
      // printf("Wrap! Now at: %ld\n",curcnt);
      curcnt = b->GetTotalLen();
      wrap = 1;
    }
    
    // Update peaks & averages to current position    
    while (lastcnt < curcnt) {
      sample_t *sptr[2];
      sample_t s_l, s_r;
      if (stereo) {
	mi->GetFragment(&sptr[0],&sptr[1]);	
	s_l = *sptr[0];
	s_r = *sptr[1];
      } else {
	mi->GetFragment(&sptr[0],0);	
	s_l = *sptr[0];
	s_r = 0;
      }

      // If chunkcnt is -1, we have temporarily stopped until a wrap
      if (go) {
	if (s_l > runmax)
	  runmax = s_l;
	if (s_l < runmin)
	  runmin = s_l;

	if (stereo) {
	  if (s_r > runmax)
	    runmax = s_r;
	  if (s_r < runmin)
	    runmin = s_r;

	  runtally += (fabs(s_l)+fabs(s_r))/2;
	}
	else
	  runtally += fabs(s_l);
	  
	chunkcnt++;
	
	if (chunkcnt >= pa->chunksize) {
	  // One chunk done
	  sample_t peak = runmax-runmin,
	    avg = (sample_t) (runtally/pa->chunksize);
	  
	  if (peaksi->PutFragment(&peak,0) || 
	      avgsi->PutFragment(&avg,0)) {
	    // Peaks should not be wrapping before main buf, stop!
	    //
	    // Note that this does happen!!
	    go = 0;
	  }
	  else {
	    peaksi->NextFragment();
	    avgsi->NextFragment();
	    chunkcnt = 0;
	  }
	  
	  runtally = 0;
	  runmax = 0;
	  runmin = 0;
	}
      }
      
      mi->NextFragment();
      lastcnt++;
    }
    
    if (wrap) {
      /* printf("Main wrap!: MI: %ld PI: %ld AI: %ld\n",
	 mi->GetTotalLength2Cur(),
	 peaksi->GetTotalLength2Cur(),
	 avgsi->GetTotalLength2Cur()); */
      mi->Zero();
      peaksi->Zero();
      avgsi->Zero();
      lastcnt = 0; // Wrap to beginning, and do the samples there
      go = 1;
      
      chunkcnt = 0;
      runtally = 0;
      runmax = 0;
      runmin = 0;
    }
  } while (wrap);

  return 0;
};

void StripeBlockManager::Setup() {
  // Does this block have BED_MarkerPoints?
  mp = (BED_MarkerPoints *)(b->GetExtendedData(T_BED_MarkerPoints));
  if (mp == 0) {
    // No marker block  
    b->AddExtendedData(mp = new BED_MarkerPoints());
  }
};

int StripeBlockManager::Manage() {
  // This is called whenever a time marker should be striped
  // to the block

  char wrap = 0;
  do {
    if (wrap)
      wrap = 2; // Special case if we just wrapped and are on a 2nd pass
    else
      wrap = 0;

    // Get current iterated offset into block
    nframes_t curcnt = i->GetTotalLength2Cur();

    if (curcnt < lastcnt) {
      // We have a wrap condition!- loop
      // printf("Wrap! Now at: %ld\n",curcnt);
      curcnt = b->GetTotalLen();
      wrap = 1;
    }

    // Scan through time markers for a good place to insert a new one 
    TimeMarker *cur = mp->markers,
      *prev = 0;
    if (wrap == 2) {
      // On 2nd pass, delete markers at 0.
      wrap = 0;
      while (cur != 0 && cur->markofs < lastcnt) {
	prev = cur;
	cur = cur->next;
      }
    } else {
      while (cur != 0 && cur->markofs <= lastcnt) {
	prev = cur;
	cur = cur->next;
      }
    }

    // Delete any markers between last position & current
    while (cur != 0 && cur->markofs <= curcnt) {
      TimeMarker *tmp = cur->next;
      cur->RTDelete();
      
      cur = tmp;
      if (prev != 0)
	prev->next = cur;
      else
	mp->markers = cur;
    } 

    if (!wrap) {
      // Now insert a new marker in this position
      // Use realtime new method- no problem
      TimeMarker *nw = (TimeMarker *) pre_tm->RTNew();
      if (nw == 0) {
	printf("StripeBlockManager: ERROR: No free TimeMarkers\n");
	return 0;
      }
      nw->markofs = curcnt;
      nw->next = cur;
      if (prev != 0)
	prev->next = nw;
      else
	mp->markers = nw;
      
      // Update counters
      lastcnt = curcnt;
    }
    else
      lastcnt = 0; // Start again from beginning to curcnt
  } while (wrap);

  return 0;
};

int DelayProcessorCallManager::Manage() {
  // OK, we've been triggered!
  //printf("Call processor %ld- %d\n",(long) p, data);

  // Call processor
  data = p->ProcessorCall(trigger,data);

  if (data == -1) {
    // Now remove this manager
    return 1; 
  } else {
    // Save this manager
    return 0;
  }
};

BlockManager::BlockManager (Fweelin *app) : 
  manageblocks(0), himanageblocks(0), threadgo(1), app(app) {
  pre_growchain = new PreallocatedType(app->getMMG(),
				       ::new GrowChainManager(),
				       sizeof(GrowChainManager));
  pre_peaksavgs = new PreallocatedType(app->getMMG(),
				       ::new PeaksAvgsManager(),
				       sizeof(PeaksAvgsManager));
  pre_hipri = new PreallocatedType(app->getMMG(),
				   ::new HiPriManagedChain(),
				   sizeof(HiPriManagedChain));
  pre_stripeblock = new PreallocatedType(app->getMMG(),
					 ::new StripeBlockManager(),
					 sizeof(StripeBlockManager));
  pre_delayprocessorcall = 
    new PreallocatedType(app->getMMG(),
			 ::new DelayProcessorCallManager(),
			 sizeof(DelayProcessorCallManager));

  // Start a block managing thread
  int ret = pthread_create(&manage_thread,
			   0,
			   run_manage_thread,
			   static_cast<void *>(this));
  if (ret != 0) {
    printf("(blockmanager) pthread_create failed, exiting");
    exit(1);
  }
  
  struct sched_param schp;
  memset(&schp, 0, sizeof(schp));
  // Manage thread calls delete- can't be SCHED_FIFO
  schp.sched_priority = sched_get_priority_max(SCHED_OTHER);
  //  schp.sched_priority = sched_get_priority_min(SCHED_FIFO);
  if (pthread_setschedparam(manage_thread, SCHED_OTHER, &schp) != 0) {    
    printf("BLOCK: Can't set realtime thread, will use nonRT!\n");
  }
}

BlockManager::~BlockManager () {
  // Terminate the management thread
  threadgo = 0;
  pthread_join(manage_thread,0);

  delete pre_growchain;
  delete pre_peaksavgs;
  delete pre_hipri;
  delete pre_stripeblock;
  delete pre_delayprocessorcall;
}

// Turns on automatic allocation of new blocks at the end of the
// specified chain. We work in conjunction with the specified iterator, 
// so that when the iterator approaches the end of the block chain, 
// the chain grows automatically
void BlockManager::GrowChainOn (AudioBlock *b, AudioBlockIterator *i) {
  // Check if the block is already being watched
  DelManager(&manageblocks,b,T_MC_GrowChain);
  // Tell the manage thread to grow this chain
  GrowChainManager *nw = (GrowChainManager *) pre_growchain->RTNew();
  nw->b = b;
  nw->i = i;
  AddManager(&manageblocks,nw);
}

void BlockManager::GrowChainOff (AudioBlock *b) {
  DelManager(&manageblocks,b,T_MC_GrowChain);
}

// Turns on computation of running sample peaks and averages
// for the specified Block & Iterator
// We compute as the currenty iterated position advances
PeaksAvgsManager *BlockManager::PeakAvgOn (AudioBlock *b, 
					   AudioBlockIterator *i,
					   char grow) {
  // Check if the block is already being watched
  DelManager(&manageblocks,b,T_MC_PeaksAvgs);
  // Tell the manage thread to compute peaks & averages
  PeaksAvgsManager *nw = (PeaksAvgsManager *) pre_peaksavgs->RTNew();
  nw->bmg = this;
  nw->b = b;
  nw->i = i;
  nw->grow = grow;
  nw->Setup();
  AddManager(&manageblocks,nw);
  
  return nw;
}

void BlockManager::PeakAvgOff (AudioBlock *b) {
  DelManager(&manageblocks,b,T_MC_PeaksAvgs);
}

void BlockManager::StripeBlockOn (void *trigger, AudioBlock *b, 
				  AudioBlockIterator *i) {
  // Check if the block is already being watched
  DelHiManager(&himanageblocks,b,T_MC_StripeBlock,trigger);
  // Tell the manage thread to stripe beats on blocks
  StripeBlockManager *nw = (StripeBlockManager *) pre_stripeblock->RTNew();
  nw->pre_tm = app->getPRE_TIMEMARKER();
  nw->b = b;
  nw->i = i;
  nw->trigger = trigger;
  nw->Setup();
  AddManager((ManagedChain**) &himanageblocks,nw);
}

void BlockManager::StripeBlockOff (void *trigger, AudioBlock *b) {
  DelHiManager(&himanageblocks,b,T_MC_StripeBlock,trigger);
}

// Removes all delayprocessorcallmanagers on the given trigger
// that call the given processor
void BlockManager::FlushDelayProcessorCall (void *trigger, Processor *p) {
  HiPriManagedChain *cur = himanageblocks,
    *prev = 0; 
  while (cur != 0) {
    if (cur->GetType() == T_MC_DelayProcessorCall &&
	(cur->trigger == 0 || cur->trigger == trigger) &&
	((DelayProcessorCallManager *) cur)->p == p) {
      // Remove chain
      HiPriManagedChain *tmp = (HiPriManagedChain *) cur->next;
      if (prev != 0) 
	prev->next = tmp;
      else 
	himanageblocks = tmp;
      cur->RTDelete();
      
      // Next chain
      cur = himanageblocks; // Start again!
    } else {
      // Next chain
      prev = cur;
      cur = (HiPriManagedChain *) cur->next;
    }
  }
};

void BlockManager::AddDelayProcessorCall (void *trigger, Processor *p, 
					  int data) {
  DelayProcessorCallManager *nw = (DelayProcessorCallManager *)
    pre_delayprocessorcall->RTNew();
  nw->trigger = trigger;
  nw->p = p;
  nw->data = data;
  AddManager((ManagedChain **) &himanageblocks,nw);
};

void BlockManager::DelDelayProcessorCall (DelayProcessorCallManager *m) {
  DelManager((ManagedChain **) &himanageblocks,m);
};

// Generic delete/add functions for managers (not hipri)
void BlockManager::DelManager (ManagedChain *m) {
  DelManager(&manageblocks,m);
};
void BlockManager::AddManager (ManagedChain *nw) {
  AddManager(&manageblocks,nw);
};

void BlockManager::DelManager (ManagedChain **first, ManagedChain *m) {
  ManagedChain *cur = *first,
    *prev = 0;
  
  // Search for manager 'm' in our list
  while (cur != 0 && cur != m) {
    prev = cur;
    cur = cur->next;
  }
  
  if (cur != 0) {
    // Got it, unlink!
    if (prev != 0) 
      prev->next = cur->next;
    else 
      *first = cur->next;
    cur->RTDelete();
  }
};

void BlockManager::RefDeleted (ManagedChain **first, void *ref) {
  ManagedChain *cur = *first,
    *prev = 0; 
  while (cur != 0) {
    if (cur->RefDeleted(ref)) {
      //printf("RefDeleted delete loose reference!\n");

      // Remove chain
      ManagedChain *tmp = cur->next;
      if (prev != 0) 
	prev->next = tmp;
      else 
	*first = tmp;
      cur->RTDelete();
      
      // Next chain
      cur = *first; // Start again!
    } else {
      // Next chain
      prev = cur;
      cur = cur->next;
    }
  }
}

// Notify all Managers that the object pointed to has been deleted-
// To avoid broken dependencies
void BlockManager::RefDeleted (void *ref) {
  RefDeleted(&manageblocks,ref);
  RefDeleted((ManagedChain **) &himanageblocks,ref);
}

// Activate a hipriority trigger- all hiprimanagedchains with
// specified trigger pointer will have manage() method called
// Safe to call in realtime!
void BlockManager::HiPriTrigger (void *trigger) {
  HiPriManagedChain *cur = himanageblocks,
    *prev = 0;
  
  // Count the managers- this is necessary
  // Because new managers can be added during the Trigger process
  // causing infinite loops
  long mgrcnt = 0;
  while (cur != 0) {
    mgrcnt++;
    cur = (HiPriManagedChain *) cur->next;
  }
  
  //printf("HIPRITRIG: %ld\n", mgrcnt);
  
  cur = himanageblocks;
  long curcnt = 0;
  while (curcnt < mgrcnt && cur != 0) {
    char advance = 1;
    
    if (cur->trigger == trigger || cur->trigger == 0)
      // Ok, right trigger or no trigger specified, call 'em!
      if (cur->Manage()) {
	// Remove chain
	HiPriManagedChain *tmp = (HiPriManagedChain *) cur->next;
	if (prev != 0) 
	  prev->next = tmp;
	else 
	  himanageblocks = tmp;
	cur->RTDelete();
	
	// Next chain
	cur = tmp;
	curcnt++;
	
	advance = 0;
      }
    
    if (advance) {
      // Next chain
      prev = cur;
      cur = (HiPriManagedChain *) cur->next;
      curcnt++;
    }
  }
}

// Returns the 1st chain manager associated with block b
// that has type t
ManagedChain *BlockManager::GetBlockManager(AudioBlock *o, 
					    ManagedChainType t) {
  ManagedChain *cur = manageblocks;
  
  // Search for block 'o' && type t in our list
  while (cur != 0 && (cur->b != o || cur->GetType() != t))
    cur = cur->next;
  
  return cur;
};

void BlockManager::AddManager (ManagedChain **first, ManagedChain *nw) {
  ManagedChain *cur = *first;
  if (cur == 0)
    *first = nw; // That was easy, now we have 1 item
  else {
    while (cur->next != 0)
      cur = cur->next;
    cur->next = nw; // Link up the last item to new1
  }
}

// Delete a managed chain for block o and manager type t
// If t is T_MC_None, removes the first managed chain for 'o' 
// of any type
void BlockManager::DelManager (ManagedChain **first, AudioBlock *o,
			       ManagedChainType t) {
  ManagedChain *cur = *first,
    *prev = 0;
  
  // Search for block 'o' in our list of type 't'
  while (cur != 0 && (cur->b != o || (cur->GetType() != t &&
				      t != T_MC_None))) {
    prev = cur;
    cur = cur->next;
  }
  
  if (cur != 0) {
    // Got it, unlink!
    if (prev != 0) 
      prev->next = cur->next;
    else 
      *first = cur->next;
    cur->RTDelete();
  }
}

// Delete a hiprimanaged chain for block o and manager type t
// with specified trigger.
// If t is T_MC_None, removes the first managed chain for 'o' 
// with specified trigger, of any type
void BlockManager::DelHiManager (HiPriManagedChain **first, 
				 AudioBlock *o,
				 ManagedChainType t, void *trigger) {
  HiPriManagedChain *cur = *first,
    *prev = 0;
  
  // Search for block 'o' in our list of type 't'
  while (cur != 0 && (cur->b != o || cur->trigger != trigger ||
		      (cur->GetType() != t && t != T_MC_None))) {
    prev = cur;
    cur = (HiPriManagedChain *) cur->next;
  }
  
  if (cur != 0) {
    // Got it, unlink!
    if (prev != 0) 
      prev->next = cur->next;
    else 
      *first = (HiPriManagedChain *) cur->next;
    cur->RTDelete();
  }
}

void *BlockManager::run_manage_thread (void *ptr) {
  BlockManager *inst = static_cast<BlockManager *>(ptr);
  
  while (inst->threadgo) {
    // Manage the blocks we have
    ManagedChain *cur = inst->manageblocks,
      *prev = 0;
    while (cur != 0) {
      if (cur->Manage()) {
	// Remove chain
	ManagedChain *tmp = cur->next;
	if (prev != 0) 
	  prev->next = tmp;
	else 
	  inst->manageblocks = tmp;
	//printf("end mgr\n");
	cur->RTDelete();
	
	// We have to start again, because manageblocks list may
	// have changed-- for ex, in destructor of manager 'cur'
	// may call DelManager
	cur = inst->manageblocks;
      }
      else {
	// Next chain
	prev = cur;
	cur = cur->next;
      }
    }

    // Produce status report?
    FloConfig *fs = inst->app->getCFG();
    if (fs->status_report == FS_REPORT_BLOCKMANAGER) {
      fs->status_report++;

      printf("BLOCKMANAGER REPORT:\n");
      ManagedChain *cur = inst->manageblocks;
      while (cur != 0) {
	printf(" bmg mgr: type(%d)\n",cur->GetType());
	cur = cur->next;
      }

      cur = inst->himanageblocks;
      while (cur != 0) {
	printf(" bmg HiPrimgr: type(%d)\n",cur->GetType());
	cur = cur->next;
      }
    }

    // 10 ms delay between management tasks
    usleep(10000);
  }

  // Delete all blockmanagers now
  ManagedChain *cur = inst->manageblocks;
  while (cur != 0) {
    ManagedChain *tmp = cur->next;
    cur->RTDelete();
    cur = tmp;
  }
  cur = inst->himanageblocks;
  while (cur != 0) {
    ManagedChain *tmp = cur->next;
    cur->RTDelete();
    cur = tmp;
  }

  return 0;
}
