#ifndef __FWEELIN_CORE_H
#define __FWEELIN_CORE_H

/*
   ********
   Art is what's going on right under our noses.
   An artist is just one who listens to it.
   ********
*/

#include <openssl/md5.h>

#include "fweelin_midiio.h"
#include "fweelin_videoio.h"
#include "fweelin_keyio.h"
#include "fweelin_audioio.h"

#include "fweelin_core_dsp.h"
#include "fweelin_block.h"
#include "fweelin_event.h"
#include "fweelin_config.h"
#include "fweelin_browser.h"

class KeyIO;
class EventManager;
class BlockManager;
class RootProcessor;
class RecordProcessor;
class TriggerMap;
class AudioBlock;
class AudioBlockIterator;
class Loop;
class Pulse;
class BED_MarkerPoints;
class VorbisStreamer;
class PreallocatedType;
class AudioBuffers;
class InputSettings;
class Browser;

#if USE_FLUIDSYNTH
class FluidSynthProcessor;
#endif

// ****************** CORE CLASSES

// Base class for all saveable types of objects
#define SAVEABLE_HASH_LENGTH MD5_DIGEST_LENGTH
#define GET_SAVEABLE_HASH_TEXT(s) \
  char hashtext[SAVEABLE_HASH_LENGTH*2+1]; \
  { \
    char *ptr = hashtext; \
    for (int i = 0; i < SAVEABLE_HASH_LENGTH; i++, ptr += 2) \
      sprintf(ptr,"%02X",s[i]); \
    *ptr = '\0'; \
  }

enum SaveStatus {
  NO_SAVE,
  SAVE_DONE
}; 

class Saveable {
 public:
  Saveable() : savestatus(NO_SAVE) {};

  virtual void Save(Fweelin *app) = 0; // Save the object

  inline unsigned char *GetSaveHash() { return savehash; };
  inline SaveStatus GetSaveStatus() { return savestatus; };
  inline void SetSaveStatus(SaveStatus s) { savestatus = s; };
  inline int SetSaveableHashFromText(char *stext) {
    int slen = strlen(stext);
    if (slen != SAVEABLE_HASH_LENGTH*2) {
      printf("DISK: Invalid MD5 hash '%s'\n",stext);
      return 1;
    }
    else {
      char htmp[3];
      htmp[2] = '\0';
      
      for (int i = 0, j = 0; i < slen; i += 2, j++) {
	htmp[0] = stext[i];
	htmp[1] = stext[i+1];
	int sh_tmp;
	sscanf(htmp,"%X",&sh_tmp);
	savehash[j] = (unsigned char) sh_tmp;
      }
      
      SetSaveStatus(SAVE_DONE);
      return 0;
    }
  };
  
 protected:

  // When we save an object, we first compute the MD5 hash for that object
  // and put that in the filename for the object. Other objects can refer to
  // the object by hash. 
  SaveStatus savestatus;
  unsigned char savehash[SAVEABLE_HASH_LENGTH+1];
};

// Loop 
// -Wraps around a list of audio blocks
// -Stores basic parameters for a loop, such as volumes
class Loop : public Saveable {
  friend class LoopManager;

public:
  const static float MIN_VOL = 0.01;

  Loop (AudioBlock *blocks, Pulse *pulse, float quant, float vol,
	long nbeats) : 
    blocks(blocks), pulse(pulse), quant(quant), 
    vol(vol), dvol(1.0), nbeats(nbeats) {};

  // Save loop
  virtual void Save(Fweelin *app);

  // RT update volume
  inline void UpdateVolume() {
    if (dvol != 1.0) {
      // Apply delta
      if (dvol > 1.0 && vol < MIN_VOL)
	vol = MIN_VOL;
      vol *=dvol;
    }
  };

  // Perhaps store the length of the loop...
  // Because right now GetTotalLength is a calculation
  // in AudioBlock

  AudioBlock *blocks; // Chain of blocks that form this loop
  
  // Time pulse to which this loop is quantized
  // And quantization factor.
  Pulse *pulse;
  float quant;

  float vol,   // Volume multiplier
    dvol;      // Rate of volume change

  long nbeats; // Number of beats in this loop
};

// This class manages sets of loops, allowing them to be triggered..
// like a keymap
class TriggerMap : public Saveable {
public:
  TriggerMap (int mapsize) : mapsize(mapsize), lastupdate(0) {
    // Map is an array of pointers to different loops
    map = new Loop *[mapsize];
    memset(map,0,mapsize * sizeof(Loop *));
  };

  virtual ~TriggerMap () {
    delete[] map;
  }

  // Save scene
  virtual void Save(Fweelin *app);

  // Go save- when all loops are saved, this part saves the XML data
  // for the scene
  void GoSave(Fweelin *app);

  inline void SetMap (int index, Loop *smp) {
    if (index < 0 || index >= mapsize) {
      printf("SetMap: Invalid loop index!\n");
    }
    else {
      map[index] = smp;
      TouchMap();
    }
  }

  inline Loop *GetMap (int index) {
    if (index < 0 || index >= mapsize) {
      printf("GetMap: Invalid loop index!\n");
      return 0;
    }
    else
      return map[index];
  }

  inline int SearchMap (Loop *l) {
    for (int i = 0; i < mapsize; i++)
      if (map[i] == l)
	return i;
    return -1;
  };

  inline int GetMapSize() { return mapsize; }
  inline double GetLastUpdate() { return lastupdate; }
  inline void TouchMap() {
    lastupdate = mygettime();
    SetSaveStatus(NO_SAVE);
  };

protected:

  int mapsize;
  Loop **map;

  // Time of last update of the map- used for rebuilding tables & such
  double lastupdate;
};

// Status types for loops
enum LoopStatus {
  T_LS_Off,

  T_LS_Recording,
  T_LS_Overdubbing,
  T_LS_Playing
};

class LoopBrowserItem : public BrowserItem {
 public:
  LoopBrowserItem(time_t time, char *name, char *fn) : BrowserItem(name),
    time(time) {
    if (fn == 0)
      filename = 0;
    else {
      // Copy filename
      filename = new char[strlen(fn)+1];
      strcpy(filename,fn);
      // Remove extension
      char *ext_ptr = strrchr(filename,'.');
      if (ext_ptr != 0)
	*ext_ptr = '\0';
    }
  };
  virtual ~LoopBrowserItem() {
    if (filename != 0)
      delete[] filename;
  };

  virtual int Compare(BrowserItem *second) { 
    if (second->GetType() == B_Loop)
      return (signed int) time-((LoopBrowserItem *) second)->time;
    else
      return 0;
  };

  virtual BrowserItemType GetType() { return B_Loop; };
  
  time_t time;
  char *filename;
};

class SceneBrowserItem : public BrowserItem {
 public:
  SceneBrowserItem(time_t time, char *name, char *fn) : BrowserItem(name),
    time(time) {
    if (fn == 0)
      filename = 0;
    else {
      // Copy filename
      filename = new char[strlen(fn)+1];
      strcpy(filename,fn);
      // Remove extension
      char *ext_ptr = strrchr(filename,'.');
      if (ext_ptr != 0)
	*ext_ptr = '\0';
    }
  };
  virtual ~SceneBrowserItem() {
    if (filename != 0)
      delete[] filename;
  };

  virtual int Compare(BrowserItem *second) { 
    if (second->GetType() == B_Scene)
      return (signed int) time-((SceneBrowserItem *) second)->time;
    else 
      return 0;
  };

  virtual BrowserItemType GetType() { return B_Scene; };
  
  time_t time;
  char *filename;
};

// LoopManager contains all loops, and wraps up recording, playing, and
// other RT & non-RT processing on loops
class LoopManager : public EventListener, public AutoWriteControl, 
		    public AutoReadControl, public BrowserCallback {
  friend class Loop;

public:
  LoopManager (Fweelin *app);
  virtual ~LoopManager();

  virtual void ReceiveEvent(Event *ev, EventProducer *from);

  virtual void ItemBrowsed(BrowserItem *i);
  virtual void ItemSelected(BrowserItem *i);

  // Populate the loop/scene browser with any loops/scenes on disk
  void SetupLoopBrowser();
  void SetupSceneBrowser();

  // Get length returns the length of any loop on the specified index
  nframes_t GetLength(int index);

  // Get length returns the length of any loop on the specified index
  // Rounded to its currently quantized length
  // Or 0 if the loop has no pulse
  nframes_t GetRoundedLength(int index);

  // Returns from 0.0-1.0 how far through a block chain with specified index
  // we are
  float GetPos(int index);

  // Get current # of samples into block chain with given index
  nframes_t GetCurCnt(int index);

  // Returns nonzero if there is a processor at the specified mapindex
  int IsActive(int index) { return (plist[index] != 0); }

  // Sets trigger volume on specified index
  // If index is not playing, activates the index
  void SetTriggerVol(int index, float vol);

  // Gets triggered volume on specified index
  // If index is not playing, returns 0
  float GetTriggerVol(int index);

  // Returns the status at the specified index
  inline LoopStatus GetStatus(int index) { return status[index]; };

  inline Processor *GetProcessor(int index) { return plist[index]; };

  // Returns a loop with the specified index, if one exists
  Loop *GetSlot(int index);

  void AdjustNewLoopVolume(float adjust) {
    newloopvol += adjust;
    if (newloopvol < 0.0)
      newloopvol = 0.0;
  }
  float GetNewLoopVolume() { return newloopvol; }

  void AdjustOutputVolume(float adjust);
  void SetOutputVolume(float set);
  float GetOutputVolume();

  void AdjustInputVolume(float adjust);
  void SetInputVolume(float set);
  float GetInputVolume();

  void SetLoopVolume(int index, float val);
  float GetLoopVolume(int index);
  void AdjustLoopVolume(int index, float adjust);
  void SetLoopdVolume(int index, float val);
  float GetLoopdVolume(int index);

  void SelectPulse(int pulseindex);

  // Create a time pulse around the specified index
  // The length of the loop on the specified index becomes
  // a time constant around which other loops center themselves
  // subdivide the length of the loop by subdivide to get the core pulse
  void CreatePulse(int index, int pulseindex, int sub);
  
  // Creates a pulse of the given length in the first available slot,
  // if none already exists of the right length
  Pulse *CreatePulse(nframes_t len);

  // Taps a pulse- starting at the downbeat- if newlen is nonzero, the pulse's
  // length is adjusted to reflect the length between taps- and a new pulse
  // is created if none exists
  void TapPulse(int pulseindex, char newlen);

  void SwitchMetronome(int pulseindex, char active);

  inline Pulse *GetCurPulse() {
    if (curpulseindex >= 0) 
      return pulses[curpulseindex];
    else
      return 0;
  };

  inline int GetCurPulseIndex() { return curpulseindex; };
  inline Pulse *GetPulseByIndex(int pulseindex) { return pulses[pulseindex]; };

  // Deletes the specified pulse, stopping striping
  void DeletePulse(int pulseindex);

  // Gets the pulse to which the loop on index is attached
  Pulse *GetPulse(int index);

  inline void SetSubdivide(int sub) {
    //printf("Set subdivide: %d\n", sub);
    subdivide = sub;
  }
  inline int GetSubdivide() { return subdivide; };

  // Move the loop at specified index to another index
  // only works if target index is empty
  // returns 1 if success
  int MoveLoop (int src, int tgt);

  // Delete the loop at the specified index..
  void DeleteLoop (int index);

  // Trigger the loop at index within the map
  // The exact behavior varies depending on what is already happening with
  // this loop and the settings passed- see .fweelin.rc
  void Activate (int index, float vol = 1.0, nframes_t ofs = 0, 
		 char overdub = 0, float od_feedback = 0.5);

  void Deactivate (int index);

  // Saves the loop with given index
  void SaveLoop (int index);
  // Load loop from disk into the given index
  void LoadLoop(char *filename, int index, float vol = 1.0);

  // Saves a scene of all loops
  void SaveScene();
  // Load scene from disk
  void LoadScene(char *filename);

  // We receive calls periodically for saving of loops-
  // here, we return blocks to save from loops which need saving
  virtual void GetWriteBlock(FILE **out, AudioBlock **b, 
			     AudioBlockIterator **i, nframes_t *len);

  // We receive calls periodically for loading of loops-
  virtual void GetReadBlock(FILE **in);
  virtual void ReadComplete(AudioBlock *b);

  // Check if the needs_saving map is up to date, rebuild if needed.
  void CheckSaveMap();

  // Adds the loop/scene with given filename to the browser br
  void AddLoopToBrowser(Browser *br, char *filename);
  void AddSceneToBrowser(Browser *br, char *filename);

  inline void SetAutoLoopSaving(char save) { autosave = save; };
  void AddToSaveQueue(Event *ev);

  Event *savequeue,          // Loop/scene save queue
    *loadqueue;              // Loop/scene load queue
  int loadloopid;            // Index where to load new loops from disk
  double needs_saving_stamp; // Timestamp from trigger map from which 
                             // needs_saving was built

  int *lastrecidx;   // List of last indexes recorded to
  int numloops,      // Total number of loops in map
    numrecordingloops; // Number of loops currently recording in map
  
 protected:

  // Adds the given loop to the list of loops to save
  void AddLoopToSaveQueue(Loop *l);
  // Adds the given loop to the list of loops to load
  void AddLoopToLoadQueue(char *filename, int index, float vol);

  // Saves loop XML data & prepares to save loop audio
  void SetupSaveLoop(Loop *l, int l_idx, FILE **out, AudioBlock **b, 
		     AudioBlockIterator **i, nframes_t *len);
  // Loads loop XML data & prepares to load loop audio-
  // returns nonzero on error
  int SetupLoadLoop(FILE **in, Loop **new_loop, int l_idx, float l_vol,
		    char *l_filename);

  // Setup time marker striping on audio memory when a new
  // pulse is selected
  void StripePulseOn(Pulse *pulse);
  
  // Deactivate time marker striping on audio memory for the
  // given pulse
  void StripePulseOff(Pulse *pulse);

  // Turn on/off auto loop saving
  char autosave; // Autosave loops?

  // Core Fweelin app on which this player is working on 
  Fweelin *app;

  // We also need to keep track of record/playprocessors attached to each
  // index
  Processor **plist;  // Pointer to a processor for each index
  // *** Replace this with well designed event queue system
  int *waitactivate;  // For each index, are we waiting to activate?
  float *waitactivate_vol;
  char *waitactivate_od;
  float *waitactivate_od_fb;
  LoopStatus *status; // For each index, what's the status?

  // Block managers that load/save loops
  BlockReadManager *bread;
  BlockWriteManager *bwrite;

  // Initial volume of new loops
  float newloopvol;

  // Subdivision for creating new pulses
  int subdivide;
  // Index of last deactivated trigger index.. used when creating new pulses
  int lastindex;
  // Current pulse (index) for new loops.. will be quantized to this pulse
  int curpulseindex;
  // List of possible pulses
  Pulse *pulses[MAX_PULSES];
};

// This is Fweelin
class Fweelin : public EventProducer {
 public:

  Fweelin() : 

#if USE_FLUIDSYNTH
    fluidp(0), 
#endif

    mmg(0), bmg(0), emg(0), rp(0), tmap(0), 
    loopmgr(0), browsers(0), abufs(0), iset(0), audio(0), midi(0), keys(0), 
    vid(0), scope(0), scope_len(0), audiomem(0), amrec(0),  
    running(0) {};
  ~Fweelin() {};

  char IsRunning() { return running; };

  // Setup
  int setup();
  // Start
  int go();

  void ToggleDiskOutput();
  void FlushWriteName() { strcpy(writename,""); };

  inline nframes_t getBUFSZ() { return fragmentsize; };
  inline sample_t *getSCOPE() { return scope; };
  inline nframes_t getSCOPELEN() { return scope_len; };

#if USE_FLUIDSYNTH
  inline FluidSynthProcessor *getFLUIDP() { return fluidp; };
  FluidSynthProcessor *fluidp;
#endif

  inline AudioBlock *getAUDIOMEM() { return audiomem; };
  AudioBlockIterator *getAUDIOMEMI();
  inline RecordProcessor *getAMREC() { return amrec; };
  AudioBlock *getAMPEAKS();
  AudioBlock *getAMAVGS();
  AudioBlockIterator *getAMPEAKSI();
  AudioBlockIterator *getAMAVGSI();
  BED_MarkerPoints *getAMPEAKSPULSE();

  inline MemoryManager *getMMG() { return mmg; };
  inline BlockManager *getBMG() { return bmg; };
  inline EventManager *getEMG() { return emg; };
  inline RootProcessor *getRP() { return rp; };
  inline VorbisStreamer *getSTREAMER() { return vs; };
  inline TriggerMap *getTMAP() { return tmap; };
  inline LoopManager *getLOOPMGR() { return loopmgr; };
  
  inline AudioIO *getAUDIO() { return audio; };
  inline MidiIO *getMIDI() { return midi; };
  inline VideoIO *getVIDEO() { return vid; };
  inline KeyIO *getKEYS() { return keys; };

  inline FloConfig *getCFG() { return cfg; };

  inline char *getOUTPUTNAME() { return writename; };

  inline PreallocatedType *getPRE_EXTRACHANNEL() { return pre_extrachannel; };
  inline PreallocatedType *getPRE_AUDIOBLOCK() { return pre_audioblock; };
  inline PreallocatedType *getPRE_TIMEMARKER() { return pre_timemarker; };

  inline AudioBuffers *getABUFS() { return abufs; };
  inline InputSettings *getISET() { return iset; };

  inline Browser *getBROWSER(BrowserItemType b) { 
    if (b >= 0 && b < B_Last)
      return browsers[b]; 
    else
      return 0;
  }

 private:
  MemoryManager *mmg;
  BlockManager *bmg;
  EventManager *emg;
  RootProcessor *rp;
  TriggerMap *tmap;
  LoopManager *loopmgr;

  Browser **browsers;

  Browser *GetBrowserFromConfig(BrowserItemType b);

  // ****************** PREALLOCATED TYPE MANAGERS
  PreallocatedType *pre_audioblock,
    *pre_extrachannel,
    *pre_timemarker;

  // ****************** VORBIS DISK STREAMER
  VorbisStreamer *vs;
  int writenum; // Number of ogg file currently being written
  char writename[FWEELIN_OUTNAME_LEN], // Name of ogg file
    timingname[FWEELIN_OUTNAME_LEN]; // Name of timing stripe file

  // ****************** SYSTEM LEVEL AUDIO  
  // Audio buffers
  AudioBuffers *abufs;
  // Input settings
  InputSettings *iset;
  // Audio interface
  AudioIO *audio;

  // ****************** MIDI
  MidiIO *midi;
  
  // ****************** KEYBOARD
  KeyIO *keys;
  
  // ****************** VIDEO
  VideoIO *vid;

  // Audio buffer size
  nframes_t fragmentsize;

  // Buffers for visual sample scope
  sample_t *scope;
  nframes_t scope_len;

  // Audio memory & the processor which records to it
  AudioBlock *audiomem;
  RecordProcessor *amrec;

  // Control settings 
  FloConfig *cfg;
  
  char running; // Nonzero if FW is fully started
};

#endif
